/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef C2S_H__
#define C2S_H__

#include "componentbase.h"
#include "glooxd.h"
#include "semafore.h"

#include <gloox/connectionhandler.h>
#include <gloox/gloox.h>
#include <gloox/mutex.h>

#include <string>
#include <list>

namespace gloox
{
  class ConnectionBase;
  class ConnectionTCPServer;
  class JID;
}

namespace glooxd
{

  class Client;
  class ConfigManager;
  class Router;
  class SM;

  /**
   * @brief The base class for the Client-to-Server (C2S) component.
   *
   * You can handle any number of domains on any number of interfaces/ports with
   * a single C2S instance.
   *
   * @author Jakob Schroeter <js@camaya.net>
   * @since 0.1
   */
  class GLOOXD_API C2S : public gloox::ConnectionHandler, ComponentBase
  {

    friend class Client;

    public:
      /**
       * Instantiates a new C2S object.
       * @param router The local Router.
       * @param cm The local ConfigManager.
       * @param logInstance The LogSink to use for logging.
       * @param sm The local session manager's name. Default: sm.
       * @param threading Whether this C2S instance runs in its own
       * thread. Default: false.
       * @param name An internal name by which this c2s is known.
       * Default: c2s.
       */
      C2S( Router& router, ConfigManager& cm, const gloox::LogSink& logInstance,
           SM& sm, bool threading = false,
           const gloox::JID& name = gloox::JID( "c2s." + GLOOXD_DOMAIN )
         );

      /**
       * Virtual destructor.
       */
      virtual ~C2S();

      /**
       * Starts processing incoming events. Blocks if the @p threading
       * parameter to the constructor is @b true.
       */
      void run();

      /**
       * Registers a new StreamFeature derived object. A copy
       * will be used by any new client connections.
       * @param feature A new stream feature.
       * @note For a list of stream features supported by default see
       * derived classes of StreamFeature. These include at least:
       * STARTTLS, SASL Auth (PLAIN only), Stream Compression, IQ Bind, IQ Session.
       */
      void registerStreamFeature( StreamFeature* feature )
      {
        m_features.push_back( feature );
      }

      // reimplemented from gloox::ConnectionHandler
      virtual void handleIncomingConnection( gloox::ConnectionBase* server,
                                             gloox::ConnectionBase* connection );

      // reimplemented from TagHandler
      virtual void handleIncomingTag( gloox::Tag* tag );

      // reimplemented from TagHandler
      virtual void handleOutgoingTag( gloox::Tag* tag );

      // reimplemented from ConfigHandler
      virtual bool handleC2SInterfaceAdded( const std::string& interface,
                                            int port );

      // reimplemented from ConfigHandler
      virtual bool handleC2SInterfaceRemoved( const std::string& interface,
                                              int port );
      // reimplemented from ConfigHandler
      virtual void handleDomainAdded( const std::string& domain );

      // reimplemented from ConfigHandler
      virtual void handleDomainRemoved( const std::string& domain );

      // reimplemented from ConfigHandler
      virtual void handleShutdown();

    private:
      /**
       * This function is used to notify C2S of the disconnect of a
       * client.
       * @param client The Client instance that got disconnected.
       */
      void disconnect( Client* client );

      /**
       * Use this periodically to receive data from the underlying client connections.
       * @param The timeout to use for select in microseconds. Default of -1 means blocking.
       * @return The state of the connection.
       */
      gloox::ConnectionError recv( int timeout = -1 );

      /**
       * Use this function to put the C2S into 'receive mode', i.e. this function returns only
       * when the connection is terminated.
       * @return Returns a value indicating the disconnect reason.
       */
      gloox::ConnectionError receive();

      void handleTag( gloox::Tag* tag );

      typedef std::list<gloox::ConnectionTCPServer*> ServerList;
      typedef std::list<Client*> ClientList;

      ServerList m_servers;
      ClientList m_clients;
      ClientList m_obsoleteClients;
      FeatureList m_features;

      gloox::util::Mutex m_clientMutex;
      gloox::util::Mutex m_obsoleteClientMutex;
      gloox::util::Mutex m_serverMutex;

      SM& m_sm;

      gloox::TagList m_queue;
      glooxd::util::Semaphore m_queueSemaphore;
      gloox::util::Mutex m_queueMutex;

      bool m_threading;
      bool m_running;

  };
}

#endif // C2S_H__
