/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef CLIENT_H__
#define CLIENT_H__

#include "glooxd.h"
#include "semafore.h"

#include <gloox/connectiondatahandler.h>
#include <gloox/jid.h>
#include <gloox/gloox.h>
#include <gloox/mutex.h>
#include <gloox/parser.h>
#include <gloox/taghandler.h>

namespace gloox
{
  class ConnectionBase;
  class IQ;
  class LogSink;
  class TLSBase;
}

namespace glooxd
{

  class AuthenticationHandler;
  class C2S;
  class ConfigManager;

  /**
   * @brief A class that abstracts a single client connection. Used by C2S.
   *
   * @author Jakob Schroeter <js@camaya.net>
   */
  class GLOOXD_API Client : public gloox::ConnectionDataHandler, gloox::TagHandler
  {
    public:
      /**
       * Creates a new Client object. It handles all the aspects of
       * a single client connection (login, etc).
       * @param c2s The parent C2S.
       * @param cm The local ConfigManager.
       * @param connection The (usually) TCP connection used for
       * communication with the client.
       * @param logInstance The LogSink to use for logging.
       */
      Client( C2S& parent, ConfigManager& cm, gloox::ConnectionBase* connection,
              const gloox::LogSink& logInstance );

      /**
       * Virtual destructor.
       */
      virtual ~Client();

      /**
       * Starts processing incoming events. Blocks if the @p threading
       * parameter to the constructor is @b true.
       * @note currently used only internally.
       */
      void run();

      /**
       * Sets the supported stream features.
       * @param features A list of supported stream features.
       */
      void setFeatures( const FeatureList& features ) { m_features = features; }

      /**
       * Use this periodically to receive data from the underlying client connection.
       * @param timeout The timeout to use for select in microseconds. Default of -1 means blocking.
       * @return The state of the connection.
       */
      gloox::ConnectionError recv( int timeout = -1 );

      /**
       * Sends the given Tag to the remote end.
       * @param tag The tag to send.
       */
      void send( gloox::Tag* tag );

      /**
       * Sends the given XML string to the remote end.
       * @param xml The XML to send.
       */
      void send( const std::string& xml );

      /**
       * Sends the given IQ to the remote end.
       * @param iq The IQ to send.
       */
      void send( const gloox::IQ& iq );

      /**
       * Sends a stream error and disconnects the client connection.
       * @param e The reason for the disconnect.
       * @param se The stream error to send. If this is
       * @p StreamErrorUndefined, no stream error will be sent.
       * @param text An optional string describing the error.
       */
      void disconnect( gloox::ConnectionError e,
                       gloox::StreamError se = gloox::StreamErrorUndefined,
                       const std::string& text = gloox::EmptyString );

      /**
       * Returns the current transport connection.
       * @return The current transport connection.
       */
      gloox::ConnectionBase* connection() { return m_connection; }

      /**
       * Returns the LogSink instance for this Client and all related objects.
       * @return The LogSink instance used in the current Client instance.
       */
      const gloox::LogSink& logInstance() { return m_logInstance; }

      /**
       * Returns the Client's JID.
       * @return The client's JID.
       */
      const gloox::JID& jid() const { return m_self; }

      /**
       * Used by the authentication routines, this function sets the
       * domain part of the remote client's authorized JID.
       * This is used if the authorization ID supplied has a different
       * server part than the stream greeting's 'to' attribute.
       * @param server The domain/server part of the JID.
       */
      void setServer( const std::string& server )
      {
        m_self.setServer( server );
      }

      /**
       * Used by the authentication routines, this function sets the
       * username the remote client authenticated as.
       * @param username The username.
       */
      void setUsername( const std::string& username )
      {
        m_self.setUsername( username );
      }

      /**
       * Used by the authentication routines, this function sets the
       * resource the remote client bound or that got assigned by the SM.
       * @param resource The resource.
       */
      void setResource( const std::string& resource )
      {
        m_self.setResource( resource );
      }

      /**
       * Used by StreamFeature-derived classes, this function
       * finishes processing of a given stream feature by
       * adding (bitwise OR) a State to the current stream state,
       * disabling the exclusive Handler (usually the
       * StreamFeature that calls finishStreamfeature()),
       * and optionally pluging a new connection on top of
       * the current connection chain.
       * @param state The StreamState to add.
       * @param conn An optional connection to use.
       */
      void finishStreamFeature( StreamState state,
                                gloox::ConnectionBase* conn = 0 );

      /**
       * Hands the tag over to the router.
       * @param tag A Tag to route.
       */
      void routeTag( gloox::Tag* tag );


      /**
       * Returns the JID by which this Client is addressable.
       * @return The JID by which this Client is addressable.
       *
       */
      const gloox::JID name() const;

      /**
       * Returns the client's internal unique ID.
       * @return The client's internal unique ID.
       */
      const std::string& id() const { return m_id; }

      /**
       * Tags passed to this function should be addressed to the Client
       * instance and will be processed locally (i.e. they will not be
       * sent to the remote client).
       * @param tag A Tag to process.
       */
      void handleIncomingTag( gloox::Tag* tag );

      /**
       * Tags passed to this function should be addressed to the
       * remote client and will be sent there (notwithstanding
       * a set exclusive handler). They will not be otherwise
       * processed locally.
       * @param tag A Tag to forward.
       */
      void handleOutgoingTag( gloox::Tag* tag );


      // re-implemented from gloox::ConnectionDataHandler
      virtual void handleReceivedData( const gloox::ConnectionBase* connection,
                                       const std::string& data );

      // re-implemented from gloox::ConnectionDataHandler
      virtual void handleConnect( const gloox::ConnectionBase* connection );

      // re-implemented from gloox::ConnectionDataHandler
      virtual void handleDisconnect( const gloox::ConnectionBase* connection,
                                     gloox::ConnectionError reason );

      // re-implemented from gloox::TagHandler
      virtual void handleTag( gloox::Tag* tag );

    private:
      void sendStreamFeatures( const std::string& domain );
      void handleStreamGreeting( const gloox::Tag* tag );
      bool handlesDomain( const std::string& domain );
      /**
       * Stamps a 'from' to the Tag and hands it over to routeTag().
       * @param tag A Tag to route.
       */
      void stampTag( gloox::Tag* tag );

      C2S& m_parent;
      ConfigManager& m_cm;

      gloox::ConnectionBase* m_connection;
      const gloox::LogSink& m_logInstance;
      gloox::Parser m_parser;

      FeatureList m_features;

      StreamFeature* m_exclusiveHandler;
      gloox::StringList m_domains;

      std::string m_id;

      int m_state;

      gloox::JID m_self;

      gloox::TagList m_incomingQueue;
      gloox::TagList m_outgoingQueue;
      gloox::util::Mutex m_incomingQueueMutex;
      gloox::util::Mutex m_outgoingQueueMutex;
      glooxd::util::Semaphore m_queueSemaphore;

      bool m_threading;
      bool m_fakePresence;
//       bool m_cancel;

  };

}

#endif // CLIENT_H__
