/*
  Copyright (c) 2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#include "confighandler.h"
#include "configmanager.h"
#include "servereventhandler.h"

#include <gloox/logsink.h>
#include <gloox/mutexguard.h>
#include <gloox/util.h>

namespace glooxd
{

  // --- ConfigManager::Interface ---
  bool ConfigManager::Interface::hasDomain( const std::string& domain )
  {
    gloox::StringList::const_iterator it = m_domains.begin();
    for( ; it != m_domains.end() && (*it) != domain; ++it )
      ;
    return it != m_domains.end();
  }
// --- ~ConfigManager::Interface ---

  ConfigManager::ConfigManager( const gloox::LogSink& logInstance )
    : m_logInstance( logInstance )
  {
  }

  ConfigManager::~ConfigManager()
  {
    gloox::util::clearList( m_c2sInterfaces );
  }

  void ConfigManager::registerConfigHandler( ConfigHandler* ch )
  {
    if( !ch )
      return;

    m_configHandlersMutex.lock();
    m_configHandlers.push_back( ch );
    m_configHandlersMutex.unlock();
  }

  void ConfigManager::registerServerEventHandler( ServerEventHandler* seh )
  {
    if( !seh )
      return;

    m_eventHandlersMutex.lock();
    m_eventHandlers.push_back( seh );
    m_eventHandlersMutex.unlock();
  }

  void ConfigManager::bindDomain( const std::string& domain, const std::string& interface,
                                  unsigned short port )
  {
    if( domain.empty() )
      return;

    m_configHandlersMutex.lock();
    gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleDomainAdded, domain );
    m_configHandlersMutex.unlock();

    gloox::util::MutexGuard mg( m_c2sInterfacesMutex );
    InterfaceList::const_iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      if( (*it)->interface() == interface && (*it)->port() == port )
      {
        (*it)->addDomain( domain );
        return;
      }
    }
  }

  void ConfigManager::unbindDomain( const std::string& domain, const std::string& interface,
                                    unsigned short port )
  {
    if( domain.empty() )
      return;

    m_configHandlersMutex.lock();
    gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleDomainRemoved, domain );
    m_configHandlersMutex.unlock();

    gloox::util::MutexGuard mg( m_c2sInterfacesMutex );
    InterfaceList::const_iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      if( (*it)->interface() == interface && (*it)->port() == port )
      {
        (*it)->removeDomain( domain );
        return;
      }
    }
  }

  bool ConfigManager::addC2SInterface( const std::string& interface,
                                       unsigned short port )
  {
    if( findInterface( interface, port ) )
      return false;

    m_c2sInterfacesMutex.lock();
    m_c2sInterfaces.push_back( new Interface( interface, port ) );
    m_c2sInterfacesMutex.unlock();

    m_configHandlersMutex.lock();
    gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleC2SInterfaceAdded, interface, port );
    m_configHandlersMutex.unlock();
    return true;
  }

  bool ConfigManager::removeC2SInterface( const std::string& interface,
                                          unsigned short port )
  {
    gloox::util::MutexGuard mg( m_c2sInterfacesMutex );
    InterfaceList::iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      if( (*it)->interface() == interface && (*it)->port() == port )
      {
        delete (*it);
        m_c2sInterfaces.erase( it );
        gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleC2SInterfaceRemoved, interface, port );
        return true;
      }
    }

    return false;
  }

  bool ConfigManager::findInterface( const std::string& interface,
                                     unsigned short port )
  {
    gloox::util::MutexGuard mg( m_c2sInterfacesMutex );
    InterfaceList::const_iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      if( (*it)->interface() == interface && (*it)->port() == port )
        return true;
    }
    return false;
  }

  void ConfigManager::rebind()
  {
    gloox::util::MutexGuard mg( m_configHandlersMutex );
    InterfaceList::const_iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      const std::string& interface = (*it)->interface();
      int port = (*it)->port();
      gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleC2SInterfaceRemoved, interface, port );
      gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleC2SInterfaceAdded, interface, port );
    }
  }

  void ConfigManager::shutdown()
  {
    gloox::util::MutexGuard mg( m_configHandlersMutex );
    gloox::util::ForEach( m_configHandlers, &ConfigHandler::handleShutdown );
  }

  bool ConfigManager::checkDomain( const std::string& domain, const std::string& ip,
                                   int port )
  {
#ifdef DEBUG
    printf( "checking %s on %s:%d\n", domain.c_str(), ip.c_str(), port );
#endif
    gloox::util::MutexGuard mg( m_c2sInterfacesMutex );
    InterfaceList::const_iterator it = m_c2sInterfaces.begin();
    for( ; it != m_c2sInterfaces.end(); ++it )
    {
      if( ( (*it)->interface() == gloox::EmptyString
              || (*it)->interface() == ip )
          && (*it)->port() == port
          && (*it)->hasDomain( domain ) )
        return true;
    }
    return false;
  }

  void ConfigManager::handleC2SListening( bool yes )
  {
    gloox::util::MutexGuard mg( m_eventHandlersMutex );
    gloox::util::ForEach( m_eventHandlers, &ServerEventHandler::handleC2SListening, yes );
  }

  void ConfigManager::handleClientConnected( const gloox::JID jid, const std::string localIP, int localPort,
                                             const std::string remoteIP, int remotePort )
  {
    gloox::util::MutexGuard mg( m_eventHandlersMutex );
    ServerEventHandlerList::const_iterator it = m_eventHandlers.begin();
    for( ; it != m_eventHandlers.end(); ++it )
      (*it)->handleClientConnected( jid, localIP, localPort, remoteIP, remotePort );
  }

  void ConfigManager::handleClientDisconnected( const gloox::JID jid, gloox::ConnectionError e,
                                                gloox::StreamError se )
  {
    gloox::util::MutexGuard mg( m_eventHandlersMutex );
    gloox::util::ForEach( m_eventHandlers, &ServerEventHandler::handleClientDisconnected, jid, e, se );
  }

}
