/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#include "simpleserver.h"

#include "../c2s.h"
#include "../certificategenerator.h"
#include "../discomanager.h"
#include "../configmanager.h"
#include "../featurestarttls.h"
#include "../featuresasl.h"
#include "../featurecompression.h"
#include "../featureresourcebind.h"
#include "../featuresession.h"
#include "../messagemanager.h"
#include "../presencemanager.h"
#include "../rostermanager.h"
#include "../router.h"
#include "../sm.h"
#include "../vcardmanager.h"

#include <gloox/gloox.h>

using namespace glooxd;

SimpleServer::SimpleServer()
{
  m_log.registerLogHandler( gloox::LogLevelDebug,
                    gloox::LogAreaAll
                        | static_cast<gloox::LogArea>( LogAreaC2S )
                        | static_cast<gloox::LogArea>( LogAreaClient )
                        | static_cast<gloox::LogArea>( LogAreaSM )
                        | static_cast<gloox::LogArea>( LogAreaDM )
                        | static_cast<gloox::LogArea>( LogAreaRouter ),
                        this );

  ConfigManager cm( m_log );
  // Do not add domains just yet. They are pushed to components
  // only once, so add them when all components are in place.

  Router r( cm, m_log );

  SM sm( r, cm, m_log, this );
  sm.addPlugin( new RosterManager( sm, r, this ) );
  sm.addPlugin( new PresenceManager( sm, r, this ) );
  sm.addPlugin( new MessageManager( sm, r, this ) );
  sm.addPlugin( new VCardManager( sm, r ) );
  sm.addPlugin( new DiscoManager( sm, r ) );
//   sm.addPlugin( new OfflineMessage( sm, r, this ) );
//   sm.addPlugin( new PrivateXML( sm, r, this ) );
//   sm.addPlugin( new Privacy( sm, r, this ) );
// etc

  C2S c2s( r, cm, m_log, sm );
  c2s.registerStreamFeature( new FeatureSTARTTLS( true, this ) );
  c2s.registerStreamFeature( new FeatureSASL( this ) );
  c2s.registerStreamFeature( new FeatureCompression() );
  c2s.registerStreamFeature( new FeatureResourceBind( sm ) );
  c2s.registerStreamFeature( new FeatureSession() );
//   c2s.registerClientHandler( this );

  // It's ok to add domains and interfaces now.
  cm.addC2SInterface( gloox::EmptyString, 6222 );
  cm.bindDomain( "gloox.im", gloox::EmptyString, 6222 );
  cm.bindDomain( "glooxd.im", gloox::EmptyString, 6222 );
  cm.bindDomain( "camaya.im", gloox::EmptyString, 6222 );

  while( true )
  {
    r.run();
    sm.run();
    c2s.run();
  }
}

SimpleServer::~SimpleServer()
{
}

bool SimpleServer::handleUserRequest( const gloox::JID& bare )
{
  printf( "user account checked: %s\n", bare.full().c_str() );
  return true;
}

bool SimpleServer::handleSASLPlain( const std::string& authzid,
                                    const std::string& authcid,
                                    const std::string& password )
{
  printf( "SASL PLAIN requested:\n    authzid: %s,\n    authcid: %s,\n    password: %s\n", authzid.c_str(), authcid.c_str(), password.c_str() );
  // Check credentials here. Return true if ok, false otherwise.
  return true;
}

const Roster SimpleServer::getRoster( const gloox::JID& /*jid*/ )
{
  glooxd::Roster rm;
  rm.insert( std::make_pair( "foo@glooxd.im", gloox::S10nFrom ) );
  rm.insert( std::make_pair( "bar@glooxd.im", gloox::S10nTo ) );
  rm.insert( std::make_pair( "js@glooxd.im", gloox::S10nBoth ) );
  rm.insert( std::make_pair( "hurkhurk@glooxd.im", gloox::S10nBoth ) );
  return rm;
}

const std::string SimpleServer::handleKeyRequest( const std::string& domain )
{
  printf( "private key requested for %s\n", domain.c_str() );
  return "/tmp/server.key";
}

const std::string SimpleServer::handleCertificateRequest( const std::string& domain )
{
  printf( "cert requested for %s\n", domain.c_str() );

  // certificates should be cached, and possibly created upfront anyway
  const std::string cert = "/tmp/" + domain + ".crt";
  if( !util::generateCertificate( domain, "/tmp/server.key",
                                  cert, 1024, 365, 1 ) )
  {
    printf( "generating cert for %s failed\n", domain.c_str() );
    return gloox::EmptyString;
  }

  return cert;
}

void SimpleServer::handleLog( gloox::LogLevel level, gloox::LogArea area, const std::string& message )
{
  printf("log: level: %d, area: %d, %s\n", level, area, message.c_str() );
}

int main( int, char** )
{
  SimpleServer s;
  return 0;
}
