/*
  Copyright (c) 2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/

#include "authenticationhandler.h"
#include "client.h"
#include "featuresasl.h"

#include <gloox/adhoc.h>
#include <gloox/base64.h>
#include <gloox/dataform.h>
#include <gloox/gloox.h>
#include <gloox/iq.h>

namespace glooxd
{

  FeatureSASL::FeatureSASL( AuthenticationHandler* authHandler )
    : m_state( SASLCanceled ), m_authHandler( authHandler )
  {
  }

  const std::string& FeatureSASL::filterString() const
  {
    static const std::string filter = "/auth[@xmlns='" + gloox::XMLNS_STREAM_SASL + "']";
    return filter;
  }

  gloox::Tag* FeatureSASL::tag( int state, const std::string& /*domain*/ )
  {
    gloox::Tag* t = 0;

    if( ( state & StateAuthentication ) != StateAuthentication
        && ( state & StateTLS ) == StateTLS )
    {
      t = new gloox::Tag( "mechanisms" );
      t->setXmlns( gloox::XMLNS_STREAM_SASL );
      new gloox::Tag( t, "mechanism", "PLAIN" );

      m_state = SASLOffered;
    }

    return t;
  }

  void FeatureSASL::handleTag( gloox::Tag* _tag )
  {
    if( !_tag )
      return;

    const gloox::Tag* auth = _tag->findTag( "/auth/[@xmlns='" + gloox::XMLNS_STREAM_SASL + "'][@mechanism='PLAIN']" );
    if( m_state == SASLOffered && auth )
    {
      decodeSASLPLAIN( auth->cdata() );
      return;
    }

    m_state = SASLCanceled;
    m_parent->disconnect( gloox::ConnAuthenticationFailed );
  }

  void FeatureSASL::decodeSASLPLAIN( const std::string& cdata )
  {
    const std::string decoded = gloox::Base64::decode64( cdata );

    std::string::size_type pos = decoded.find( '\0', 0 );
    if( pos == std::string::npos )
    {
      m_parent->disconnect( gloox::ConnAuthenticationFailed );
      return;
    }

    const gloox::JID authzid( decoded.substr( 0, pos ) );

    std::string::size_type pos2 = decoded.find( '\0', pos + 1 );
    if( pos2 == std::string::npos )
    {
      m_parent->disconnect( gloox::ConnAuthenticationFailed);
      return;
    }

    const std::string& authcid = decoded.substr( pos + 1, pos2 - 1 );
    const std::string& password = decoded.substr( pos2 + 1 );

    m_state = SASLChecking;
    bool success = false;
    if( m_authHandler )
      success = m_authHandler->handleSASLPlain( authzid.bare(),
                                                authcid, password );

    gloox::Tag* s = new gloox::Tag( success ? "success" : "failure" );
    s->setXmlns( gloox::XMLNS_STREAM_SASL );
    m_parent->send( s );

    if( success )
    {
      m_parent->finishStreamFeature( StateAuthentication );
      m_parent->finishStreamFeature( StateReset );
      if( authzid )
      {
        const gloox::JID a( authzid );
        m_parent->setUsername( a.username() );
        m_parent->setServer( a.server() );
      }
      else
        m_parent->setUsername( authcid );

      m_state = SASLEnabled;
    }
    else
    {
      m_state = SASLCanceled;
      m_parent->disconnect( gloox::ConnAuthenticationFailed );
    }
  }

}
