/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


/*! @mainpage glooxd API Documentation
 *
 * @section contents Contents
 * @ref intro_sec <br>
 * @ref c2s_sec <br>
 * @ref auth_sec <br>
 * @ref roster_sec <br>
 * @ref basicserver_sec <br>
 * @ref knownissues_sec <br>
 * <br>
 *
 * @section intro_sec Introduction
 *
 * The design of <a href='http://camaya.net/glooxd'>glooxd</a> closely follows the design of its
 * underlying library, <a href='http://camaya.net/gloox'>gloox</a>, and tries to re-use as much
 * as possible from it.
 *
 * Amongst others, the design features:
 * @li Stream encryption
 * @li Stream compression
 * @li Pluggable stream features
 * @li Pluggable SASL authentication modules
 * @li Pluggable client connections
 * @li Support for seaparately threading each client connection
 *
 * @section c2s_sec Client-2-Server connections
 *
 * Client-2-Server connections are handled by @link glooxd::C2S C2S @endlink instances.
 * You need at least one @link glooxd::C2S C2S @endlink instance, which can handle as
 * many domains as you like. You can have as many @link glooxd::C2S C2S @endlink instances as you like.
 *
 * @section auth_sec Authentication
 *
 * When a client connects, it will be able to use only SASL PLAIN for authentication.
 * Use the @link glooxd::AuthenticationHandler AuthenticationHandler @endlink to check credential supplied by the client.
 *
 * @section roster_sec Roster Management
 *
 * There is no Roster Management by means of XMPP implemented yet. However,
 * by using the @link glooxd::RosterProvider RosterProvider @endlink,
 * you can send a custom roster to clients.
 *
 * @section basicserver_sec Basic Server
 *
 * To set up a basic server, use code similar to this:
 * @code
 * Router r;
 * SM sm( this, r.logInstance() );
 * C2S c2s( &r, sm, this, r.logInstance() );
 * Interface* i = new Interface(); // all local interfaces, port 5222
 * c2s.addDomain( "mydomain.com", i );
 * c2s.receive();
 * @endcode
 *
 * @section knownissues_sec Known Issues and Shortcomings
 *
 * @li Server-2-server connections are not supported,
 * @li Subscriptions are not handled at all,
 * @li Roster management (add, remove) is not available,
 * @li Presence priority is not taken into account at all,
 * @li Messages to bare JIDs are sent to all connected resources,
 * @li Automatic resource allocation is limited: The assigned resource
 * is always 'foo'.
 */

#ifndef GLOOXD_H__
#define GLOOXD_H__

#include "macros.h"

#include <gloox/gloox.h>

#include <string>
#include <list>

namespace gloox
{
  class Tag;
}

/**
 * @brief The principal namespace for glooxd, an embeddable XMPP
 * server.
 *
 * @author Jakob Schroeter <js@camaya.net>
 */
namespace glooxd
{

  class StreamFeature;

  /** Supported stream version (major). */
  GLOOXD_API extern const std::string XMLNS_SERVER;

  /** Supported stream version (major). */
  GLOOXD_API extern const std::string XMPP_STREAM_VERSION_MAJOR;

  /** Supported stream version (minor). */
  GLOOXD_API extern const std::string XMPP_STREAM_VERSION_MINOR;

  /** glooxd private namespace */
  GLOOXD_API extern const std::string XMLNS_GLOOXD;

  /** glooxd version */
  GLOOXD_API extern const std::string GLOOXD_VERSION;

  /** glooxd internal domain (address suffix) */
  GLOOXD_API extern const std::string GLOOXD_DOMAIN;

  /**
   * glooxd log areas.
   */
  enum LogArea
  {
    LogAreaC2S                        = 0x1000000, /**< Log messages from C2S. */
    LogAreaSM                         = 0x2000000, /**< Log messages from SM. */
    LogAreaS2S                        = 0x4000000, /**< Log messages from S2S. */
    LogAreaRouter                     = 0x8000000, /**< Log messages from Router. */
    LogAreaClient                     = 0x10000000,/**< Log messages from Client. */
    LogAreaDM                         = 0x20000000 /**< Log messages from DM. */
  };

  /**
   * Describes the current stream state.
   */
  enum StreamState
  {
    StateDisconnected       =  0,   /**< No client connected. */
    StateReset              =  1,   /**< The stream has been reset, stream greeting expected */
    StateTransportConnected =  2,   /**< We have someone connected on the other end. */
    StateTLS                =  4,   /**< TLS handshake complete, stream encrypted. */
    StateAuthentication     =  8,   /**< Client authenticated. */
    StateCompression        = 16,   /**< Stream compressed. */
    StateDone               = 32,   /**< Stream initiation is done. */
    StateReallyDone         = 64    /**< Stream init really done. */
  };

  /**
   * A list of stream features.
   */
  typedef std::list<StreamFeature*> FeatureList;

  /**
   * Returns a customized error stanza depending on the given
   * parameters.
   * @param tag A Tag to reply to.
   * @param e A StanzaError to include in the reply.
   */
  gloox::Tag* returnError( const gloox::Tag* tag,
                           gloox::StanzaError e = gloox::StanzaErrorServiceUnavailable );

}

extern "C"
{
  GLOOXD_API const char* glooxd_version();
}

#endif // GLOOXD_H__
