/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef ROUTER_H__
#define ROUTER_H__

#include "confighandler.h"
#include "glooxd.h"
#include "semafore.h"
#include "taghandler.h"

#include <gloox/mutex.h>
#include <gloox/mutexguard.h>
#include <gloox/tag.h>

#include <string>

namespace gloox
{
  class LogSink;
}

namespace glooxd
{

  class ConfigManager;

  /**
   * @brief The base class for the routing component.
   *
   * @author Jakob Schroeter <js@camaya.net>
   */
  class GLOOXD_API Router : public TagHandler, public ConfigHandler
  {
    public:
      /**
       * Instantiates a new object that can be used for local (in-process) routing.
       * @param cm The local ConfigManager.
       * @param logInstance The local LogSink.
       * @param threading Whether this instance runs in its own thread.
       */
      Router( ConfigManager& cm, const gloox::LogSink& logInstance, bool threading = false );

      /**
       * Destructor.
       */
      ~Router();

      /**
       * Starts processing incoming events. Blocks if the @p threading
       * parameter to the constructor is @b true.
       */
      void run();

      /**
       * Registers a TagHandler for the given domain. Used internally
       * to route Stanzas to components.
       * @param th The TagHandler to register for the given domain.
       * @param domain The domain to route to the given TagHandler.
       * @param node The routed JID must posses a node to be routed to
       * this TagHandler.
       */
      void registerDomain( TagHandler* th, const std::string& domain, bool node );

      /**
       */
      void removeDomain( TagHandler* th, const std::string& domain, bool node );

      // reimplemented from TagHandler
      virtual void handleIncomingTag( gloox::Tag* tag );

      // reimplemented from TagHandler
      virtual void handleOutgoingTag( gloox::Tag* ) {}

      // reimplemented from ConfigHandler
      virtual void handleDomainAdded( const std::string& ) {}

      // reimplemented from ConfigHandler
      virtual void handleDomainRemoved( const std::string& ) {}

      // reimplemented from ConfigHandler
      virtual void handleShutdown();

    private:
      void handleTag( gloox::Tag* tag );
      void routeBySender( gloox::Tag* tag );
      void routeByReceipient( gloox::Tag* tag );
      bool isLocal( const std::string& domain );

      typedef std::map<std::string, TagHandler*> RoutingTable;
      RoutingTable m_domainRouting;
      RoutingTable m_userRouting;

      ConfigManager& m_cm;

      const gloox::LogSink& m_logInstance;
      gloox::util::Mutex m_domainRoutingMutex;
      gloox::util::Mutex m_userRoutingMutex;

      gloox::TagList m_queue;
      glooxd::util::Semaphore m_queueSemaphore;
      gloox::util::Mutex m_queueMutex;

      bool m_threading;

  };

}

#endif // ROUTER_H__
