/*
  Copyright (c) 2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef SERVEREVENTHANDLER_H__
#define SERVEREVENTHANDLER_H__

#include "macros.h"

#include <gloox/jid.h>

#include <string>

namespace glooxd
{

  /**
   * @brief An interface that can be re-implemented to receive
   * server events.
   *
   * @author Jakob Schroeter <js@camaya.net>
   * @since 0.2
   */
  class GLOOXD_API ServerEventHandler
  {
    public:
      /**
       * Virtual destructor.
       */
      virtual ~ServerEventHandler() {}

      /**
       * This function is called with an parameter of @b true exactly
       * @b once when one or more c2s interfaces have been added. I.e.
       * it is called when the number of listening interfaces is
       * exactly 1.
       * It is called with a parameter of @b false when interfaces have
       * been removed and the number of listening interfaces is now
       * exactly 0.
       * @param yes Whether or not there is at least one listening
       * interface.
       */
      virtual void handleC2SListening( bool yes ) { (void)yes; }

      /**
       * This function is called when a client connected, authenticated,
       * and requested creation of a session.
       * @param jid The client's full JID.
       * @param localIP The local IP the client connected to.
       * @param localPort The local port the client connected to.
       * @param remoteIP The remote IP the client connected from.
       * @param remotePort The remote port the client connected from.
       */
      virtual void handleClientConnected( const gloox::JID jid, const std::string localIP, int localPort,
                                          const std::string remoteIP, int remotePort )
      {
        (void)jid;
        (void)localIP;
        (void)localPort;
        (void)remoteIP;
        (void)remotePort;
      }

      /**
       * This function is called when a client disconnected or
       * got disconnected. It will only be called for client
       * connections that handleClientConnected() has been called for
       * previously.
       * @param jid The client's full JID.
       * @param e A connection error.
       * @param se A stanza error.
       */
      virtual void handleClientDisconnected( const gloox::JID jid, gloox::ConnectionError e,
                                             gloox::StreamError se )
      {
        (void)jid;
        (void)e;
        (void)se;
      }

  };

}

#endif // SERVEREVENTHANDLER_H__
