/*
  Copyright (c) 2008-2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef SM_H__
#define SM_H__

#include "componentbase.h"
#include "glooxd.h"

#include <gloox/gloox.h>
#include <gloox/jid.h>
#include <gloox/mutex.h>

#include <list>
#include <map>
#include <string>

namespace gloox
{
  class LogSink;
}

namespace glooxd
{

  class Client;
  class Plugin;
  class RosterProvider;

  /**
   * @brief Implements a client session manager.
   *
   * @author Jakob Schroeter <js@camaya.net>
   */
  class GLOOXD_API SM : public ComponentBase
  {
    public:
      /**
       * Creates a new SessionManager.
       * @param router The local Router.
       * @param cm The local ConfigManager.
       * @param logInstance The LogSink to use for logging.
       * @param rp A RosterProvider derived object that will provide
       * the roster.
       * @param threading Whether this instance runs in its own
       * thread.
       * @param name An internal name by which this SM is known.
       * Default: sm.
       */
      SM( Router& router, ConfigManager& cm, const gloox::LogSink& logInstance,
          RosterProvider* rp, bool threading = false,
          const gloox::JID& name = gloox::JID( "sm." + GLOOXD_DOMAIN ) );

      /**
       *  Virtual destructor.
       */
      virtual ~SM();

      /**
       * Starts processing incoming events. Blocks if the @p threading
       * parameter to the constructor is @b true.
       */
      void run();

      /**
       * Checks whether a new session for the given JID can be
       * created, and creates it if possible. A unique resource
       * identifier will be assigned if none is given.
       * @param jid The JID to create a session for. It may be
       * modified (ie. the resource may be updated).
       * @param id The client's connection's ID.
       * @return @b True if the session could be created
       * successfully, @b false otherwise.
       */
      bool createSession( gloox::JID& jid, const std::string& id );

      /**
       * Removes the session (identified by its ID).
       * @param id The ID of the session to be removed.
       * @return @b True if a session was found and removed, @b false
       * otherwise.
       */
      bool removeSession( const std::string& id );

      /**
       * Adds a protocol handling plugin to this SM instance.
       * @param p The Plugin to add.
       */
      void addPlugin( Plugin* p )
      {
        if( p )
          m_plugins.push_back( p );
      }

      /**
       * Sends broadcast and unicast presence packets to their
       * destinations.
       * @param pres The presence Tag to send.
       */
      void dispatchPresence( const gloox::Tag* pres );

      /**
       * Sends message packets to their destinations.
       * @param msg The message Tag to send.
       */
      void dispatchMessage( const gloox::Tag* msg );

      // reimplemented from TagHandler
      virtual void handleIncomingTag( gloox::Tag* tag );

      // reimplemented from TagHandler
      virtual void handleOutgoingTag( gloox::Tag* ) {}

      // reimplemented from ConfigHandler
      virtual void handleDomainAdded( const std::string& domain );

      // reimplemented from ConfigHandler
      virtual void handleDomainRemoved( const std::string& domain );

      // reimplemented from ConfigHandler
      virtual void handleShutdown();

    private:
      struct Session
      {
        std::string id;
        gloox::Tag* presence;
        int priority;
      };

      /**
       * A map of full JID/session pairs.
       */
      typedef std::map<std::string, Session> SessionList;

      typedef std::list<Plugin*> PluginList;

      void handleTag( gloox::Tag* tag );

      /**
       * Checks if the there are open sessions for the given bare JID.
       * @param jid The bare JID to look for sessions for.
       * @return A list of open sessions for that JID.
       */
      const SessionList findSessions( const gloox::JID& jid );

      SessionList m_sessions;
      PluginList m_plugins;
      RosterProvider* m_rosterProvider;

      gloox::util::Mutex m_sessionMutex;

      gloox::TagList m_queue;
      glooxd::util::Semaphore m_queueSemaphore;
      gloox::util::Mutex m_queueMutex;

      bool m_threading;

  };

}

#endif // SM_H__
