/*
  Copyright (c) 2009 by Jakob Schroeter <js@camaya.net>
  This file is part of the glooxd library. http://camaya.net/glooxd

  This software is distributed under a license. The full license
  agreement can be found in the file LICENSE in this distribution.
  This software may not be copied, modified, sold or distributed
  other than expressed in the named license agreement.

  This software is distributed without any warranty.
*/


#ifndef STREAMFEATURE_H__
#define STREAMFEATURE_H__

#include "glooxd.h"

#include <string>

namespace gloox
{
  class Tag;
}

namespace glooxd
{

  class Client;
  class TagHandler;

  /**
   * @brief This is the base class for a strean feature. It allows to
   * plug new features (and associated processing) into the server.
   *
   * @author Jakob Schroeter <js@camaya.net>
   * @since 0.2
   */
  class GLOOXD_API StreamFeature
  {
    public:
      /**
       * Constructor.
       * @param parent The Client instance to use for sending data.
       */
      StreamFeature()
        : m_parent( 0 ), m_state( 0 )
      {}

      /**
       * Virtual destructor.
       */
      virtual ~StreamFeature() {}

      /**
       * Sets the Client instance that will be used for communication.
       * You should not need to use this function manually.
       */
      void setParent( Client* parent ) { m_parent = parent; }

      /**
       * Returns an XPath expression that describes the elements that
       * are associated with this stream feature.
       * @return An XPath expression.
       */
      virtual const std::string& filterString() const = 0;

      /**
       * Returns a Tag structire for inclusion in the stream features to send to the client.
       * Features may depend on current stream state and the domain the client requested.
       * @return A feature Tag. May be 0 if there is no applicable feature for the
       * current state and domain.
       */
      virtual gloox::Tag* tag( int state, const std::string& domain ) = 0;

      /**
       * Returns a new Instance of the derived type. It should be configured identically,
       * but should not inherit any state.
       * The simplest implementation for a derived class FooFeature looks like:
       * @code
       * StreamFeature* FooFeature::newInstance() const
       * {
       *   return new FooExtension();
       * }
       * @endcode
       *
       * @return The derived feature's new instance.
       */
      virtual StreamFeature* newInstance() const = 0;

      /**
       *
       */
      virtual void handleTag( gloox::Tag* tag ) = 0;

    protected:
      Client* m_parent;
      int m_state;

  };

}

#endif // STREAMFEATURE_H__
